<?php

 /*
 * Upela connector - shipping module
 * More info on www.upela.com
 * @copyright Upela <https://www.upela.com>
 * @author Benjamin Abbou <benjamin.abbou@hexalogic.fr>
 * @version 1.0.0
 */

  define('REQUIRE_SECURE', FALSE);
  $moduleVersion = '1.0.0';
  $schemaVersion = '1.0.0';

  // include the Mage engine
  umask(0);
  
//  use Magento\Framework\App\Bootstrap;


use Magento\Framework\AppInterface;
use Magento\Catalog\Model\Product;

try {
    require __DIR__ . '/app/bootstrap.php';

//    use Magento\Catalog\Model\Product;

    $m2bootstrap = \Magento\Framework\App\Bootstrap::create(BP, $_SERVER);
    $m2objGlobal = $m2bootstrap->getObjectManager();
    $state = $m2objGlobal->get('Magento\Framework\App\State');
    $state->setAreaCode('frontend');
} catch (\Exception $e) {
    echo 'Autoload error: ' . $e->getMessage();
    exit(1);
}

  
  
  // retrieve the store code
  $storeCode = '';
  if (isset($_REQUEST['storecode'])) {
    $storeCode = $_REQUEST['storecode'];    
  }
  
  // using output buffering to get around headers that magento is setting after we've started output
  ob_start();

  header('Content-Type: text/xml;charset=utf-8');
  header('Last-Modified: ' . gmdate('D, d M Y H:i:s') . ' GMT');
  
  // HTTP/1.1
  header('Cache-Control: no-store, no-cache, must-revalidate');
  header('Cache-Control: post-check=0, pre-check=0', false);

  // HTTP/1.0
  header('Pragma: no-cache');     

  // Tests the Magento version to see if it's greater than or equal to the targetVersion
  function MagentoVersionGreaterOrEqualTo($targetVersion){
    return true;
  }

  // write xml documenta declaration
  function writeXmlDeclaration() {
    echo '<?xml version="1.0" standalone="yes" ?>';
  }
  
  function writeStartTag($tag, $attributes = null) {
    echo '<' . $tag;
    
    if ($attributes != null) {
        echo ' ';
        
        foreach ($attributes as $name => $attribValue) {
            echo $name . '="' . htmlspecialchars($attribValue) . '" ';  
        }
    }
    
    echo '>';
  }
  
  // write closing xml tag
  function writeCloseTag($tag) {
    echo '</' . $tag . '>';
  }

  // Output the given tag\value pair
  function writeElement($tag, $value) {
    writeStartTag($tag);
    echo htmlspecialchars($value);
    writeCloseTag($tag);
  }

  // Outputs the given name/value pair as an xml tag with attributes
  function writeFullElement($tag, $value, $attributes) {
    echo '<' . $tag . ' ';

    foreach ($attributes as $name => $attribValue) {
      echo $name. '="' . htmlspecialchars($attribValue) . '" '; 
    }
    echo '>';
    echo htmlspecialchars($value);
    writeCloseTag($tag);
  }
  
  // Function used to output an error and quit.
  function outputError($code, $error) {
    writeStartTag("Error");
    writeElement("Code", $code);
    writeElement("Description", $error);
    writeCloseTag("Error");
  }       

  $secure = false;
  try {
    if (isset($_SERVER['HTTPS'])) {
        $secure = ($_SERVER['HTTPS'] == 'on' || $_SERVER['HTTPS'] == '1');
    }
  } catch(Exception $e) {
  }
              
  // Open the XML output and root
  writeXmlDeclaration();
  writeStartTag('Upela', array('moduleVersion' => $moduleVersion, 'schemaVersion' => $schemaVersion));
  
// Enforse SSL
  if (!$secure && REQUIRE_SECURE) {
    outputError(10, 'A secure (https://) connection is required.');
  } else {
      // If the admin module is installed, we make use of it
      if (checkAdminLogin($m2objGlobal)) {
        $action = (isset($_REQUEST['action']) ? $_REQUEST['action'] : '');
        switch (strtolower($action)) {
                case 'getmodule': getModule(); break;
                case 'getstore': getStore($m2objGlobal); break;
                case 'getcount': getCount($m2objGlobal); break;
                case 'getorders': getOrders($m2objGlobal); break;
                case 'getstatuscodes': getStatusCodes($m2objGlobal); break;
                case 'updateorder': updateOrder($m2objGlobal); break;
                case 'gettest' : getTest(); break;
                default:
                    outputError(20, "'$action' is not supported.");
        }
      }
  }
  
  // Close the output
  writeCloseTag('Upela');
  
  function getTest(){
    echo "Test OK";
    
  }
  
  // Check to see if admin functions exist.  And if so, determine if the user
  // has access.
  function checkAdminLogin($m2obj) {
    $loginOK = false;
  
    if (isset($_REQUEST['username']) && isset($_REQUEST['password'])) {
      $username = $_REQUEST['username'];
      $password = $_REQUEST['password'];

      $user = $m2obj->create('\Magento\User\Model\User')
                    ->login($username, $password);
      
      if ($user && $user->getId()) {
        $loginOK = true;
      }
    }
  
    if (!$loginOK) {
      outputError(50, 'The username or password is incorrect.');
    }
  
    return $loginOK;
  }

  function getModule() {
    writeStartTag('Module');
    
        writeElement('Platform', 'Magento');
        writeElement('Developer', 'Interapptive, Inc. (support@interapptive.com)');
        
    writeStartTag('Capabilities');
    writeElement('DownloadStrategy', 'ByModifiedTime');
    writeFullElement('OnlineCustomerID', '', array('supported' => 'true', 'dataType' => 'numeric'));
    writeFullElement('OnlineStatus', '', array('supported' => 'true', 'dataType' => 'text', 'downloadOnly' => 'true'));
    writeFullElement('OnlineShipmentUpdate', '', array('supported' => 'false'));
    writeCloseTag('Capabilities');
    
    writeCloseTag('Module');
  }
  
  // Write store data
  function getStore($m2obj) {
    // get state name
    /*$region_model = Mage::getModel('directory/region');
    if (is_object($region_model)) {
      $state = $region_model->load(Mage::getStoreConfig('shipping/origin/region_id'));
      $stateName = $state->getDefaultName();
      $stateCode = $state->getCode();
    }*/
    
    $config = $m2obj->create('\Magento\Framework\App\Config\ScopeConfigInterface');
    
    $countryFactory = $m2obj->get('Magento\Directory\Model\CountryFactory');
    $stateCode = $config->getValue('shipping/origin/region_id');
    $country_id = $config->getValue('shipping/origin/country_id');

    $regions = $countryFactory
                    ->create()
                    ->setId($country_id)
                    ->getLoadedRegionCollection()
                    ->toOptionArray();
                    
    foreach($regions as $region){
        if($region['value'] == $stateCode){
            $stateName= $region['title'];    
            break;
        }
    }


    $name = $config->getValue('general/store_information/name');
    $phone = $config->getValue('general/store_information/phone');
    $owner = $config->getValue('trans_email/ident_general/name');
    $email = $config->getValue('trans_email/ident_general/email');
    $country = $config->getValue('shipping/origin/country_id');
    $postcode = $config->getValue('shipping/origin/postcode');
    $city = $config->getValue('shipping/origin/city');
    $address1 = $config->getValue('shipping/origin/street_line1');
    $address2 = $config->getValue('shipping/origin/street_line2');
    
    $storeManager = $m2obj->get('Magento\Store\Model\StoreManagerInterface');
    $website= $storeManager->getStore()->getUrl();
    
    writeStartTag('Store');
      writeElement('Platform', 'magento');
    writeElement('Name', $name);
    writeElement('CompanyOrOwner', $owner);
    writeElement('Email', $email);
    writeElement('Street1', $address1);
    writeElement('Street2', $address2);
    writeElement('Street3', '');
    writeElement('City', $city);
    writeElement('State', $stateName);
    writeElement('StateCode', $stateCode);
    writeElement('PostalCode', $postcode);
    writeElement('Country', $country);
    writeElement('CountryCode', $country);
    writeElement('Phone', $phone);
    writeElement('Website', $website);
    writeCloseTag('Store');
  }

  // Converts an xml datetime string to sql date time
  function toLocalSqlDate($sqlUtc) {
    $pattern = "/^(\d{4})-(\d{2})-(\d{2})\T(\d{2}):(\d{2}):(\d{2})$/i";
  
    if (preg_match($pattern, $sqlUtc, $dt)) {
      $unixUtc = gmmktime($dt[4], $dt[5], $dt[6], $dt[2], $dt[3], $dt[1]);  
      return date("Y-m-d H:i:s", $unixUtc);
    }
  
    return $sqlUtc;
  }

  // Get the count of orders greater than the start ID
  function getCount($m2obj) {
    $start = 0;

    $storeManager = $m2obj->get('Magento\Store\Model\StoreManagerInterface');

    $storeId = $storeManager->getStore()->getId();

    if (isset($_REQUEST['start'])) {
      $start = $_REQUEST['start'];
    }

    // only get orders through 2 seconds ago
    $end = date('Y-m-d H:i:s', time() - 2);
    // Convert to local SQL time
    $start = toLocalSqlDate($start);

    // Write the params for easier diagnostics
    writeStartTag('Parameters');
    writeElement('Start', $start);
    writeCloseTag('Parameters');

    $ordersColl = $m2obj->get('\Magento\Sales\Model\ResourceModel\Order\CollectionFactory');
    $orders = $ordersColl->create()->addFieldToSelect('*')->addFieldToFilter('updated_at',['gt'=> $start])->addFieldToFilter('updated_at',['lteq'=>$end])->addFieldToFilter('store_id',$storeId);

    $count = $orders->count();

    writeElement('OrderCount', $count);
  }

  // Get all orders greater than the given start id, limited by max count
  function getOrders($m2obj) {
    $storeManager = $m2obj->get('Magento\Store\Model\StoreManagerInterface');
    $storeId = $storeManager->getStore()->getId();

    
    $start = 0;
    $maxcount = 100;
  
    if (isset($_REQUEST['start'])) {
      $start = $_REQUEST['start'];
    }
  
    if (isset($_REQUEST['maxcount'])) {
      $maxcount = $_REQUEST['maxcount'];
    }
  
    // Only get orders through 2 seconds ago.
    $end = date('Y-m-d H:i:s', time() - 2);
  
    // Convert to local SQL time
    $start = toLocalSqlDate($start);
  
    // Write the params for easier diagnostics
    writeStartTag('Parameters');
    writeElement('Start', $start);
    writeElement('End', $end);
    writeElement('MaxCount', $maxcount);
    writeCloseTag('Parameters');                                    
  
    // setup the query
    $ordersColl = $m2obj->get('\Magento\Sales\Model\ResourceModel\Order\CollectionFactory');
    $orders = $ordersColl->create()->addFieldToSelect('*')->addFieldToFilter('updated_at',['gt'=> $start])->addFieldToFilter('updated_at',['lteq'=>$end])->addFieldToFilter('store_id',$storeId)->setOrder('updated_at','asc');
    
    // configure paging
    $orders->setCurPage(1)
      ->setPageSize($maxcount)
      ->loadData();
  
    writeStartTag('Orders');
  
    $lastModified = null;
    $processedIds = '';
  
    foreach ($orders as $order) {
      // keep track of the ids we've downloaded
      $lastModified = $order->getUpdatedAt();
  
      if ($processedIds != '') {
        $processedIds .= ', ';
      }
      $processedIds .= $order->getEntityId();
  
      WriteOrder($order, $m2obj);
    }
  
    writeCloseTag('Orders');
  }
  
  // Output the order as xml
  function WriteOrder($order, $m2obj) {                 
    global $secure;
    writeStartTag('Order');
  
    $incrementId = $order->getIncrementId();
    $orderPrefix = '';
    $orderPostfix = '';
    $prefixParts = preg_split('/(?=\d)/', $incrementId,2);
      
    $orderPrefix = $prefixParts[0];
  
    $postfixParts = preg_split('[\-]', $prefixParts[1],2);
    
    $incrementId = $postfixParts[0];    
    if (count($postfixParts) >= 2)
        $orderPostfix = $postfixParts[1];
    
    
    writeElement('OrderNumberPrefix', $orderPrefix);
    writeElement('OrderNumber', $incrementId);
    writeElement('OrderNumberPostfix', $orderPostfix);
    writeElement('OrderDate', FormatDate($order->getCreatedAt()));
    writeElement('LastModified', FormatDate($order->getUpdatedAt()));
    writeElement('ShippingMethod', $order->getShippingDescription());
    writeElement('StatusCode', $order->getStatus());
    writeElement('CustomerID', $order->getCustomerId());

        // check for order-level gift messages
      /*  if ($order->getGiftMessageId()) {
        $message = Mage::helper('giftmessage/message')->getGiftMessage($order->getGiftMessageId());
        $messageString = 'Gift message for '. $message['recipient']. ': '. $message['message'];
    
        writeStartTag('Notes');
        writeFullElement('Note', $messageString, array('public' => 'true'));
        writeCloseTag('Notes');
        }
*/
    $address = $order->getBillingAddress();

    writeStartTag('BillingAddress');
    writeElement('FullName', $address->getName());
    writeElement('Company', $address->getCompany());
    writeElement('Street1', $address->getStreetLine(1));
    writeElement('Street2', $address->getStreetLine(2));
    writeElement('Street3', $address->getStreetLine(3));
    writeElement('City', $address->getCity());
    writeElement('State', $address->getRegionCode());
    writeElement('StateCode', $address->getRegionCode());
    writeElement('PostalCode', $address->getPostcode());
    writeElement('Country', $address->getCountryId());
    writeElement('CountryCode', $address->getCountryId());
    writeElement('Phone', $address->getTelephone());
    writeElement('Email', $order->getCustomerEmail());
    writeCloseTag('BillingAddress');

    $billFullName = $address->getName();
    $billStreet1 = $address->getStreet(1);
    $billCity = $address->getCity();
    $billZip = $address->getPostcode();
    
    $address = $order->getShippingAddress();
    if (!$address) {
      // sometimes the shipping address isn't specified, so use billing
      $address = $order->getBillingAddress();
    }

    writeStartTag('ShippingAddress');
    writeElement('FullName', $address->getName());
    writeElement('Company', $address->getCompany());
    writeElement('Street1', $address->getStreetLine(1));
    writeElement('Street2', $address->getStreetLine(2));
    writeElement('Street3', $address->getStreetLine(3));
    writeElement('City', $address->getCity());
    writeElement('State', $address->getRegionCode());
    writeElement('StateCode', $address->getRegionCode());
    writeElement('PostalCode', $address->getPostcode());
    writeElement('Country', $address->getCountryId());
    writeElement('CountryCode', $address->getCountryId());
    writeElement('Phone', $address->getTelephone());

    // if the addressses appear to be the same, use customer email as shipping email too
    if ($address->getName() == $billFullName &&
    $address->getStreet(1) == $billStreet1 &&
    $address->getCity() == $billCity &&
    $address->getPostcode() == $billZip) {
      writeElement('Email', $order->getCustomerEmail());
    }
  
    writeCloseTag('ShippingAddress');


    $payment = $order->getPayment();

    // CC info
    if ($secure) {
      $cc_num = $payment->getCcNumber();
    } else {
      $cc_num = $payment->getCcLast4();
      if (!empty($cc_num)) {
       $cc_num = '************'.$payment->getCcLast4();
      }
    }
    $cc_year = sprintf('%02u%s', $payment->getCcExpMonth(), substr($payment->getCcExpYear(), 2)); 

    writeStartTag('Payment');

    $helper = $m2obj->get('\Magento\Payment\Helper\Data');
    writeElement('Method', $helper->getMethodInstance($payment->getMethod())->getTitle());    
    


    writeStartTag('CreditCard');
    writeElement('Type', $payment->getCcType());
    writeElement('Owner', $payment->getCcOwner());
    writeElement('Number', $cc_num);
    writeElement('Expires', $cc_year);
    writeCloseTag('CreditCard');

    writeCloseTag('Payment');

    WriteOrderItems($order->getAllItems());

    WriteOrderTotals($order);

    writeStartTag('Debug');
    writeElement('OrderID', $order->getEntityId());
    writeElement('OrderNumberPostfix', $orderPostfix);
    writeCloseTag('Debug');

    writeCloseTag('Order');
  }

  // writes a single order total
  function WriteOrderTotal($name, $value, $class, $impact = 'add') {
    if ($value > 0) {
        writeFullElement('Total', $value, array('name' => $name, 'class' => $class, 'impact' => $impact));
    }
  }
  
  // Write all totals lines for the order
  function WriteOrderTotals($order) {
    writeStartTag('Totals');
    
    writeElement('Shipping', $order->getShippingAmount());
    // Magento 1.4 started storing discounts as negative values
    if ($order->getDiscountAmount() < 0) {
      writeElement('Discounts', -1 * $order->getDiscountAmount());
    }
    writeElement('GrandTotal', $order->getGrandTotal());

    writeCloseTag('Totals');
  }

  // Gets the price of an order item
  function getCalculationPrice($item) {
        return $item->getPrice();
    return 0;
  }

  // Write XML for all products for the given order
  function WriteOrderItems($orderItems) {
    writeStartTag('Items');
    
    $parentMap = Array();
  
    // go through each item in the collection
    foreach ($orderItems as $item) {
      // keep track of item Id and types
      $parentMap[$item->getItemId()] = $item->getProductType();

      // get the sku
//print_r($item->getData());       
      $sku = $item->getSku();
      
      // weights are handled differently if the item is a bundle or part of a bundle
      $weight = $item->getWeight();
      if ($item->getIsVirtual()) {
         $weight = 0;
      }

      if ($item->getProductType() == \Magento\Catalog\Model\Product\Type::TYPE_BUNDLE) {
        $name = $item->getName(). ' (bundle)';
        $unitPrice = $item->getPrice();
      } else {
        $name = $item->getName();

        // if it's part of a bundle
        if (is_null($item->getParentItemId())) {
          $unitPrice = $item->getPrice();
        } else {
          // need to see if the parent is a bundle or not
          $isBundle = ($parentMap[$item->getParentItemId()] == \Magento\Catalog\Model\Product\Type::TYPE_BUNDLE);
          if ($isBundle) {
            // it's a bundle member - price and weight come from the bundle definition itself
            $unitPrice = 0;
            $weight = 0;
          } else {
            // don't even want to include if the parent item is anything but a bundle
            continue;
          }
        }
      }

      // Magento 1.4+ has Cost
      $unitCost = 0;
      $unitCost = $item->getBaseCost();
      
      writeStartTag('Item');
  
      writeElement('ItemID', $item->getItemId());
      writeElement('ProductID', $item->getProductId());
      writeElement('Code', $sku);
      writeElement('SKU', $sku);
      writeElement('Name', $name);
      writeElement('Quantity', (int)$item->getQtyOrdered());
      writeElement('UnitPrice', $unitPrice);
      writeElement('UnitCost', $unitCost);
  
        if (!$weight) {
            $weight = 0;
        }
      writeElement('Weight', $weight);
  
      writeStartTag('Attributes');
      $opt = $item->getProductOptions();

/*        TYPE_Configrable not in M2
      if ($item->getProductType() == Mage_Catalog_Model_Product_Type::TYPE_CONFIGURABLE) {
        if (is_array($opt)
        && isset($opt['attributes_info'])
        && is_array($opt['attributes_info'])
        && is_array($opt['info_buyRequest'])
        && is_array($opt['info_buyRequest']['super_attribute'])) {
          $attr_id = $opt['info_buyRequest']['super_attribute'];
          reset($attr_id);
          foreach ($opt['attributes_info'] as $sub) {
            writeStartTag('Attribute');
            writeElement('Name', $sub['label']);
            writeElement('Value', $sub['value']);
            writeCloseTag('Attribute');
            next($attr_id);
          }
        }
      }*/
  
      if (is_array($opt)
      && isset($opt['options'])
      && is_array($opt['options'])) {
        foreach ($opt['options'] as $sub) {
          writeStartTag('Attribute');
          writeElement('Name', $sub['label']);
          writeElement('Value', $sub['value']);
          writeCloseTag('Attribute');
        }
      }
    
        // Uncomment the following lines to include a custom product attribute in the downloaded data.
        // These will appear as Order Item Attributes in Upela.
        //$product = Mage::getModel('catalog/product');
        //$productId = $product->getIdBySku($sku);
        //$product->load($productId);
        //$value = $product->getAttributeText('attribute_code_here');
        //if ($value)
        //{
        //     // write the gift message as an attribute
        //     writeStartTag('Attribute');
        //     writeElement('Name', 'Attribute_title_here');
        //     writeelement('Value', $value);
        //     writeCloseTag('Attribute');
        //}

      writeCloseTag('Attributes');
  
      writeCloseTag('Item');
    }

    writeCloseTag('Items');
  }

  // Returns the status codes for the store
  function getStatusCodes($m2obj) {
      writeStartTag('StatusCodes');
      $statuses = $m2obj->get('Magento\Sales\Model\Order\Status')->getCollection()->toOptionArray();


//    $statuses_node = $config->getNode('global/sales/order/status');
//    print_r($statuses);
    foreach ($statuses as $status) {
      writeStartTag('StatusCode');
      writeElement('Code', $status['value']);
      writeElement('Name', $status['label']);
      writeCloseTag('StatusCode');
    }

    writeCloseTag('StatusCodes');
  }

  // Update the status of an order
  function updateOrder($m2obj) {
    // gather paramtetes
    if (!isset($_REQUEST['order'])
    //|| !isset($_REQUEST['command'])
    //|| !isset($_REQUEST['comments'])
    ) {
      outputError(40, 'Not all parameters supplied.');
      return;
    }
  
    // newer version of Upela, pull the entity id
    $orderID = (int)$_REQUEST['order'];
    //$order = Mage::getModel('sales/order')->load($orderID);
    $order = $m2obj->get('\Magento\Sales\Model\Order')->loadByIncrementId($orderID);
    
    //$command = (string) $_REQUEST['command'];
    $command = 'complete';
    $comments = isset($_REQUEST['comments']) ? $_REQUEST['comments'] : null;
    $tracking = isset($_REQUEST['tracking']) ? $_REQUEST['tracking'] : null;
    $carrierData = isset($_REQUEST['carrier']) ? $_REQUEST['carrier'] : null;

   ExecuteOrderCommand($order, $command, $comments, $carrierData, $tracking, $m2obj);
  }

  // Takes the actions necessary to get an order to Complete
  function CompleteOrder($order, $comments, $carrierData, $tracking, $m2obj) {
    // first create a shipment
    echo 'CompleteOrder';

    try  {
    if (!($order->hasShipments())){
        // Initialize the order shipment object
        $convertOrder = $m2obj->create('Magento\Sales\Model\Convert\Order');
        $shipment = $convertOrder->toShipment($order);
        
        //    print_r($shipment->getData());

        // Loop through order items
        foreach ($order->getAllItems() AS $orderItem) {
            // Check if order item has qty to ship or is virtual
            if (! $orderItem->getQtyToShip() || $orderItem->getIsVirtual()) {
                continue;
            }

            $qtyShipped = $orderItem->getQtyToShip();

            // Create shipment item with qty
            $shipmentItem = $convertOrder->itemToShipmentItem($orderItem)->setQty($qtyShipped);

            // Add shipment item to shipment
            $shipment->addItem($shipmentItem);
        }

        $shipment->register();
        if (strlen($comments) > 0) {
            $shipment->addComment($comments, false);
        }

        $shipment->getOrder()->setIsInProcess(true);
    
     

        // add tracking info if it was supplied
        if (strlen($tracking) > 0) {
//          $track = Mage::getModel('sales/order_shipment_track')->setNumber($tracking);
            $track = $m2obj->create('\Magento\Sales\Model\Order\Shipment\Track');
            $track->setNumber($tracking); 
          if (strlen($carrierData) > 0) {
            $track->setCarrierCode($carrierData);
            $track->setTitle($carrierData);
          } else {
            $track->setCarrierCode('upela');
            $track->setTitle('Upela');
          }
          //print_r($track->getData());
          $shipment->addTrack($track);
        }        

        $transactionSave = $m2obj->get('\Magento\Framework\DB\TransactionFactory');
                
        $transactionSave->create()
          ->addObject($shipment)
          ->addObject($shipment->getOrder())
          ->save();
            
         // send the email if it's requested
        if (isset($_REQUEST['sendemail']) && $_REQUEST['sendemail'] == '1') {
            $shipment->sendEmail(true);
        }
    }
    } 
    // Only one shipment can be created per order.  Subsequent attempts will throw, but we ignore the error and move on to setting up the invoice.
    catch (Exception $ex) {
    
    }
    
    // invoice the order
    $invoice = null;
    if ($order->hasInvoices()) {
      // select the last invoice to attach the note to
      $invoice = $order->getInvoiceCollection()->getLastItem();
    } else {
      // prepare a brand-new invoice
      $invoice = $order->prepareInvoice();
      //$invoice = Mage::getModel('sales/service_order', $order)->prepareInvoice();
      $invoice->register();     
    }
    
    // capture the invoice if possible
    if ($invoice->canCapture()) {
      $invoice->Capture();
    }
    
    // some magento versions prevent multiple pay() calls from have impact,
    // but others don't.  If pay is called multiple times, Order.Total Paid is off.
    if ($invoice->getState() != \Magento\Sales\Model\Order\Invoice::STATE_PAID) {
      $invoice->pay();
    }
    
    // set the comment
    if (strlen($comments) > 0) {
      $invoice->addComment($comments);
    }
    
    // save the new invoice
    $transactionSave = $m2obj->get('\Magento\Framework\DB\TransactionFactory');
    $transactionSave->create()
        ->addObject($invoice)
        ->addObject($invoice->getOrder())
        ->save();
    
    // Saving will force magento to move the state/status
    $order->save();
  }

  // Changes the status of an order 
  function ExecuteOrderCommand($order, $command, $comments, $carrierData, $tracking, $m2obj) {
    try {
      // to change statuses, we need to unhold if necessary
      if ($order->canUnhold()) {
        $order->unhold();
        $order->save();
      }


      switch (strtolower($command)) {
        case 'complete':
          CompleteOrder($order, $comments, $carrierData, $tracking, $m2obj);
          break;
        case 'cancel':
          $order->cancel();
          $order->addStatusToHistory($order->getStatus(), $comments);
          $order->save();
          break;
        case 'hold':
          $order->hold();
          $order->addStatusToHistory($order->getStatus(), $comments);
          $order->save();
          break;
        default:
          outputError(80, "Unknown order command '$command'.");
          break;
      }

      writeStartTag('Debug');
      writeElement('OrderStatus', $order->getStatus());
      writeCloseTag('Debug');
    } catch (Exception $ex) {
      outputError(90, 'Error Executing Command. '.  $ex->getMessage());
    }
  }

  // Converts a sql data string to xml date format
  function FormatDate($dateSql) {
    $pattern = '/^(\d{4})-(\d{2})-(\d{2})\s+(\d{2}):(\d{2}):(\d{2})$/i';

    if (preg_match($pattern, $dateSql, $dt)) {
      $dateUnix = mktime($dt[4], $dt[5], $dt[6], $dt[2], $dt[3], $dt[1]);
      return gmdate('Y-m-d\TH:i:s', $dateUnix);
    }

    return $dateSql;
  }

  // end output
  ob_end_flush();
?>
